/*
 * Copyright Fastly, Inc.
 * Licensed under the MIT license. See LICENSE file for details.
 */
import { spawnSync } from 'node:child_process';
import cli from '@fastly/cli';
import { makeRetryable } from './retryable.js';
export function loadApiToken(params) {
    let apiToken = null;
    let source = '';
    // Try command line arg
    if (typeof params.commandLine === 'string') {
        apiToken = params.commandLine.trim();
        source = 'commandline';
    }
    // Try env (FASTLY_API_TOKEN)
    if (apiToken == null) {
        apiToken = process.env.FASTLY_API_TOKEN || null;
        if (apiToken != null) {
            source = 'env';
        }
    }
    // Try fastly cli
    if (apiToken == null) {
        try {
            const { stdout, error } = spawnSync(cli, ['profile', 'token', '--quiet'], {
                encoding: 'utf-8',
            });
            apiToken = error ? null : stdout.trim();
        }
        catch {
            apiToken = null;
        }
        if (apiToken != null) {
            source = 'fastly-profile-token';
        }
    }
    if (apiToken == null) {
        return null;
    }
    return { apiToken, source };
}
const RETRYABLE_STATUS_CODES = [
    408, // Request Timeout
    409, // Conflict (depends)
    423, // Locked
    429, // Too Many Requests
    500, // Internal Server Error
    502, // Bad Gateway
    503, // Service Unavailable
    504, // Gateway Timeout
];
export class FetchError extends Error {
    constructor(message, status) {
        super(message);
        this.name = 'FetchError';
        this.status = status;
    }
}
function isReadableStream(data) {
    return (typeof data === 'object' &&
        data !== null &&
        typeof data.getReader === 'function' &&
        typeof data.tee === 'function');
}
export async function callFastlyApi(fastlyApiContext, endpoint, operationName, queryParams, requestInit) {
    const url = new URL(endpoint, 'https://api.fastly.com/');
    if (queryParams != null) {
        for (const [key, value] of queryParams.entries()) {
            url.searchParams.append(key, value);
        }
    }
    const headers = new Headers(requestInit?.headers);
    headers.set('Fastly-Key', fastlyApiContext.apiToken);
    const request = new Request(url, {
        ...requestInit,
        headers,
        redirect: 'error',
        ...(isReadableStream(requestInit?.body) ? { duplex: 'half' } : null),
    });
    let response;
    try {
        response = await fetch(request);
    }
    catch (err) {
        if (err instanceof TypeError) {
            throw makeRetryable(err);
        }
        else {
            throw err;
        }
    }
    if (!response.ok) {
        if (!RETRYABLE_STATUS_CODES.includes(response.status)) {
            throw new FetchError(`${operationName} failed: ${response.status}`, response.status);
        }
        throw makeRetryable(new FetchError(`Retryable ${operationName} error: ${response.status}`, response.status));
    }
    return response;
}
