/*
 * Copyright Fastly, Inc.
 * Licensed under the MIT license. See LICENSE file for details.
 */
import crypto from 'node:crypto';
import fs from 'node:fs';
import path from 'node:path';
export function enumerateFiles(opts) {
    const results = [];
    enumerateFilesWorker(results, opts.publicDirRoot, opts);
    return results;
}
function enumerateFilesWorker(results, dir, opts) {
    const entries = fs.readdirSync(dir, { withFileTypes: true });
    for (const entry of entries) {
        const { name } = entry;
        const { publicDirRoot, excludeDirs, excludeDotFiles, includeWellKnown, } = opts;
        const fullpath = path.resolve(dir, name);
        const relative = '/' + path.relative(publicDirRoot, fullpath);
        if (excludeDirs.some(excludeDir => excludeDir.test(relative))) {
            continue;
        }
        if (excludeDotFiles && name.startsWith('.')) {
            if (includeWellKnown && name === '.well-known') {
                // ok
            }
            else {
                continue;
            }
        }
        if (entry.isDirectory()) {
            enumerateFilesWorker(results, fullpath, opts);
        }
        else {
            results.push(fullpath);
        }
    }
}
export function getFileSize(filePath) {
    const stats = fs.statSync(filePath);
    return stats.size;
}
export function directoryExists(dirPath) {
    try {
        const stats = fs.statSync(dirPath);
        return stats.isDirectory();
    }
    catch (err) {
        if (err.code === 'ENOENT') {
            return false; // Path doesn't exist
        }
        throw err; // Other errors, rethrow
    }
}
export async function calculateFileSizeAndHash(filePath) {
    const stream = fs.createReadStream(filePath);
    // Use createHash, rather than subtle, since it supports streaming.
    const hash = crypto.createHash('sha256');
    let size = 0;
    return new Promise((resolve, reject) => {
        stream.on('data', chunk => {
            hash.update(chunk);
            size += chunk.length;
        });
        stream.on('end', () => {
            resolve({
                size,
                hash: hash.digest('hex'),
            });
        });
        stream.on('error', reject);
    });
}
export function rootRelative(itemPath) {
    return dotRelative(null, itemPath);
}
export function dotRelative(from, to) {
    const relPath = path.relative(from ?? path.resolve(), to);
    return relPath.startsWith('..') ? relPath : './' + relPath;
}
