/*
 * Copyright Fastly, Inc.
 * Licensed under the MIT license. See LICENSE file for details.
 */
import { callFastlyApi, FetchError } from './api-token.js';
const cache = new WeakMap();
export async function getKVStoreIdForNameMap(fastlyApiContext) {
    const cacheEntry = cache.get(fastlyApiContext);
    let kvStoreNameMap = cacheEntry?.kvStoreNameMap;
    if (kvStoreNameMap != null) {
        return kvStoreNameMap;
    }
    const kvStoreInfos = await getKVStoreInfos(fastlyApiContext);
    kvStoreNameMap = {};
    for (const { id, name } of kvStoreInfos) {
        kvStoreNameMap[name] = id;
    }
    cache.set(fastlyApiContext, { ...cacheEntry, kvStoreNameMap });
    return kvStoreNameMap;
}
export async function getKVStoreIdForName(fastlyApiContext, kvStoreName) {
    const kvStoreNameMap = await getKVStoreIdForNameMap(fastlyApiContext);
    return kvStoreNameMap[kvStoreName] ?? null;
}
function createArrayGetter() {
    return function (fn) {
        return async function (fastlyApiContext, operationName, ...args) {
            const results = [];
            let cursor = null;
            while (true) {
                const queryParams = new URLSearchParams();
                if (cursor != null) {
                    queryParams.set('cursor', cursor);
                }
                const endpoint = fn(...args);
                const response = await callFastlyApi(fastlyApiContext, endpoint, operationName, queryParams);
                const infos = await response.json();
                for (const item of infos.data) {
                    results.push(item);
                }
                if (infos.meta.next_cursor === undefined) {
                    break;
                }
                cursor = infos.meta.next_cursor;
            }
            return results;
        };
    };
}
const _getKVStoreInfos = createArrayGetter()(() => `/resources/stores/kv`);
export async function getKVStoreInfos(fastlyApiContext) {
    const cacheEntry = cache.get(fastlyApiContext);
    let kvStoreInfos = cacheEntry?.kvStoreInfos;
    if (kvStoreInfos != null) {
        return kvStoreInfos;
    }
    kvStoreInfos = await _getKVStoreInfos(fastlyApiContext, 'Listing KV Stores');
    cache.set(fastlyApiContext, { ...cacheEntry, kvStoreInfos });
    return kvStoreInfos;
}
export const _getKVStoreKeys = createArrayGetter()((kvStoreId, prefix) => {
    let endpoint = `/resources/stores/kv/${encodeURIComponent(kvStoreId)}/keys`;
    if (prefix != null) {
        endpoint += '?prefix=' + encodeURIComponent(prefix);
    }
    return endpoint;
});
export async function getKVStoreKeys(fastlyApiContext, kvStoreName, prefix) {
    const kvStoreId = await getKVStoreIdForName(fastlyApiContext, kvStoreName);
    if (kvStoreId == null) {
        return null;
    }
    return await _getKVStoreKeys(fastlyApiContext, `Listing Keys for KV Store [${kvStoreId}] ${kvStoreName}${prefix != null ? ` (prefix '${prefix}')` : ''}`, kvStoreId, prefix);
}
export async function getKvStoreEntryInfo(fastlyApiContext, kvStoreName, key) {
    return getKvStoreEntry(fastlyApiContext, kvStoreName, key, true);
}
export async function getKvStoreEntry(fastlyApiContext, kvStoreName, key, metadataOnly) {
    const kvStoreId = await getKVStoreIdForName(fastlyApiContext, kvStoreName);
    if (kvStoreId == null) {
        return null;
    }
    const endpoint = `/resources/stores/kv/${encodeURIComponent(kvStoreId)}/keys/${encodeURIComponent(key)}`;
    let response;
    try {
        response = await callFastlyApi(fastlyApiContext, endpoint, `Checking existence of [${key}]`, null, { method: metadataOnly ? 'HEAD' : 'GET' });
    }
    catch (err) {
        if (err instanceof FetchError && err.status === 404) {
            return null;
        }
        throw err;
    }
    const metadata = response.headers.get('metadata');
    const generation = response.headers.get('generation');
    return {
        metadata,
        generation,
        response,
    };
}
const encoder = new TextEncoder();
export async function kvStoreSubmitEntry(fastlyApiContext, kvStoreName, key, data, metadata) {
    const kvStoreId = await getKVStoreIdForName(fastlyApiContext, kvStoreName);
    if (kvStoreId == null) {
        throw new Error('KV Store not found');
    }
    const endpoint = `/resources/stores/kv/${encodeURIComponent(kvStoreId)}/keys/${encodeURIComponent(key)}`;
    const body = typeof data === 'string' ? encoder.encode(data) : data;
    await callFastlyApi(fastlyApiContext, endpoint, `Submitting item [${key}]`, null, {
        method: 'PUT',
        headers: {
            'content-type': 'application/octet-stream',
            ...(metadata != null ? { metadata } : null)
        },
        body,
    });
}
export async function kvStoreDeleteEntry(fastlyApiContext, kvStoreName, key) {
    const kvStoreId = await getKVStoreIdForName(fastlyApiContext, kvStoreName);
    if (kvStoreId == null) {
        throw new Error('KV Store not found');
    }
    const endpoint = `/resources/stores/kv/${encodeURIComponent(kvStoreId)}/keys/${encodeURIComponent(key)}`;
    await callFastlyApi(fastlyApiContext, endpoint, `Deleting item [${key}]`, null, {
        method: 'DELETE',
    });
}
